unit xMinRNSearchRN;
{ Feel free to use this unit as you want.
  First developped by Bernard Gisin

  This unit helps minimising function : RN -- > R, i.e. of several variables

->procedure MinimiseSimplex;
  Minimisation using the downhill simplex method in Multidimensions.
  See "Numerical recipes" Second edition Chap. 10.4,  page 408
  This unit allow to minimise a function : Rn --> R
  i.e. of several variables.
  ----------------------------------------------------------------}

INTERFACE
//"""""""

uses wGlobals2;

type
  // To have an array of vectors.
  PTypeArrayVectors = ^TypeArrayVectors;
  TypeArrayVectors = array[1..100000000] of PTypeArrayFloat;

  PMinRNSearchRN = ^TMinRNSearchRN;
  TMinRNSearchRN = object
  //=====================
  // Depending on the minimisation used, Inputs and Outputs can change
  // Must be set before calling  Minimise :
  //"""""""""""""""""""""""""""""""""""""""
  nDim : int32;
  // Dimension of the searched vector

  Func : TypeFuncRNR1Data;
  // Function to minimise

//  vSizeEdge : Float; Should be changed

  // Can be changed before calling Minimise
  //""""""""""""""""""""""""""""""""""""""""
  vSizeEdge : Float;
  // Starting size of the Simplex. (=1.0 by default)

  wOptions : word32;
  // Some options.
  // =0 by default. In this case, only the first vector must be given as a
  //    starting guess.  vSizeEdge  should be set to have a better start.
  // =1 or 3 => the starting Simplex has been given in
  //            papvecV^[ii]^[jj],  ii=1..nDim+1,  jj=1..nDim
  // =2 or 3 => The function evaluations at the starting vector has been done
  // =2      => For only the first vector
  // =3      => For all (nDim+1) vectors.
  // papvecV^[ii],  ii=1..nDim+1, define the Starting Simplex
  // papvecV^[ii] = one vector
  // pavY^[ii], ii=1..nDim+1, define the function evaluation

  vTolRel,vTolAbs : Float;
  // When the difference between functions evaluation are lower than vTolAbs,
  // Then the function is ended.

  nNbFuncEvalMax : int32;
  // Maximum number of functions evaluations accepted

  pDataFunc : pointer;
  // Pointer passed to the function when it is evaluated.

  // Variables for intermediate computations
  //"""""""""""""""""""""""""""""""""""""""""
  // For the Simplex method
  pvecGC : PTypeArrayFloat;
  // = Gravitiy Center * (nDim+1)

  pvecTry : PTypeArrayFloat;
  // Vector where the function will be evaluated

  pavY : PTypeArrayFloat;
  // memorise the function evaluations at the (nDim+1) vertices

  wSizeMemAlloc : word32;
  // Size of memory allocated for pvecTry and pvecGC

  papvecV : PTypeArrayVectors;
  // Array of vectors, each having allocated a size of wSizeMemAlloc

  nNbVecAlloc : int32;
  // Number of vectors allocated in  papvecP

  // Results :
  //""""""""""
  nExitCode : integer;
  // = 0 when the minimum has been found, i.e. the vTolEnd condition has been reached
  // = 1 when the maximum number of iterations has been reached.

  nNbFuncEval : int32;
  // number of function evaluations

  nLo, // Index of the vertex corresponding of the lowest function evaluation
  nHi  // Index of the vertex corresponding of the highest function evaluation
       : int32;

  // Used by MinimisePowell
  pvecXMin : PTypeArrayFloat;
  // On input, starting guess minimum
  // vector at which the function evaluation is the lowest

  pvecXMinLast : PTypeArrayFloat;
  // For intermediate computation, Last value of pvecXMin

  pvecDir : PTypeArrayFloat;
  // Direction in wich the minimum is searched

  pvecX : PTypeArrayFloat;
  // intermediate variable to compute the minimum of the function
  // on the line : pvecMin + lambda * pvecDir

  pvecExtrapol : PTypeArrayFloat;
  // extrapolation point from pvecXMin and pvecXMinLast

  vFXMin : Float;  // = Func(pvecXMin)

  constructor Init(nDimMaxP:int32);
  // nDimMaxP = maximum dimension.
  // usually = dimension of the vectors

  destructor Done; virtual;

  function TryNewVertex(vFac:Float):Float;
  // used by MinimiseSimplex to compute a new vertex, and evaluate the function
  // one this new point.

  procedure MinimiseSimplex; virtual;
  // Do the minimisation of the function, using the Simplex methode
  // Inputs :
  //"""""""""
  // nDim
  // func
  // papvecV^[1]^[jj],  jj=1..nDim
  // Optional inputs :
  //""""""""""""""""""
  // nNbFuncEvalMax
  // vTolRel
  // vTolAbs
  // pDataFunc
  // vSizeEdge
  // wOptions
  // Depending on wOptions, these must be given too.
  // papvecV^[ii]^[jj],  ii=1..nDim+1,  jj=1..nDim
  // papvecV^[ii] = one vector
  // pavY^[ii], ii=1..nDim+1,  the initial function evaluations
  // Outputs :
  //""""""""""
  // papvecV[nLo] = best vector found for the minimum
  // papvecV[nHi] = vector giving the highest function evaluation among the vectors kept
  // nNbFuncEval  = number of function evaluation done.
  // pavY^[nLo]   = evaluation of the function at  papvecV[nLo]
  // pavY^[nHi]   = evaluation of the function at  papvecV[nHi]

  procedure MinimisePowell; virtual;
  // Do the minimisation of the function, using the Powell's methode
  // Inputs :
  //"""""""""
  // nDim = dimension of the vectors
  // pvecXMin = starting guess minimum position
  // vSizeEdge = Distance from the starting point to begin to search the minimum
  // Func      = function to minimise
  // Optional inputs :
  //""""""""""""""""""
  // nNbFuncEvalMax
  // vTolAbs
  // vTolRel
  // wOptions
  // vFXMin = value of the function at  pvecXMin
  // Depending on wOptions, these must be given too.
  // papvecV^[ii]^[jj],  ii=1..nDim,  jj=1..nDim
  //           = Starting search directions
  // Output :
  //"""""""""
  // nExitCode=0 => found OK
  // nExitCode=1 => maximum number of iteration reached.
  // nNbFuncEval = number of functions evaluation done
  // pvecXMin = point at which the minimum function evaluation has been done
  // vFXMin = Func(pvecXMin)

  end; // TMinRNSearchRN = object

IMPLEMENTATION
//""""""""""""

Uses xMinR1SearchR1, wWindowText0, wDebug0;

var nNotUsed : integer; // only to avoid an error message while compliling in DevPascal 1.9

//****************************************************************
// TMinRNSearchRN = object
//****************************************************************

constructor TMinRNSearchRN.Init(nDimMaxP:int32);
//==============================================
var ii : int32;
begin
// +1 because the arrays begin with 0 and not 1.
wSizeMemAlloc:=(nDimMaxP+1) * SizeOf(Float);

// Used by MinimiseSimplex
GetMem(pavY, wSizeMemAlloc + SizeOf(Float));
GetMem(pvecTry, wSizeMemAlloc);
GetMem(pvecGC, wSizeMemAlloc);

// Used by MinimisePowell
GetMem(pvecXMin, wSizeMemAlloc);
GetMem(pvecXMinLast, wSizeMemAlloc);
GetMem(pvecDir, wSizeMemAlloc);
GetMem(pvecX, wSizeMemAlloc);
GetMem(pvecExtrapol, wSizeMemAlloc);

// Allocate memory for one more vector than the dimension
nNbVecAlloc :=nDimMaxP + 1;

GetMem(papvecV, nNbVecAlloc * SizeOf(pointer));
for ii:=1 to nNbVecAlloc do GetMem(papvecV^[ii], wSizeMemAlloc);

// This must still be set
nDim:=nDimMaxP;
// func = ...
// papvecV^[1]^[jj] := ...  jj:=1..nDim

// This can be changed
vTolRel   :=1e-4;
vTolAbs   :=1e-4;
nNbFuncEvalMax:=5000;
pDataFunc :=NIL;

vSizeEdge :=1.0;
wOptions  :=0;

end;  //  TMinRNSearchRN.Init

destructor TMinRNSearchRN.Done;
//=============================
var ii : int32;
begin
if (papvecV <> NIL) then begin
   for ii:=nNbVecAlloc downto 1 do FreeMem(papvecV^[ii], wSizeMemAlloc);
   FreeMem(papvecV, nNbVecAlloc * SizeOf(pointer));
   end;

// Used by MinimisePowell
FreeMem(pvecExtrapol, wSizeMemAlloc);
FreeMem(pvecX, wSizeMemAlloc);
FreeMem(pvecDir, wSizeMemAlloc);
FreeMem(pvecXMinLast, wSizeMemAlloc);
FreeMem(pvecXMin, wSizeMemAlloc);

if (pvecGC  <> NIL) then FreeMem(pvecGC, wSizeMemAlloc);
if (pvecTry <> NIL) then FreeMem(pvecTry, wSizeMemAlloc);
if (pavY    <> NIL) then FreeMem(pavY, wSizeMemAlloc + SizeOf(Float));
end;  //  TMinRNSearchRN.Done

function TMinRNSearchRN.TryNewVertex(vFac:Float):Float;
//=====================================================
var
  jj          : integer;
  vFac1,vFac2,// to compute a new trial point, candidate to be a minimum
  vYTry       // result of the function evaluation
              : Float;
begin
vFac1 :=(1.0  - vFac) / nDim;
vFac2 :=vFac1 - vFac;

// pvecTry = C_F + fac * (Hi - C_F)  where C_F = center of the oposite face
for jj:=1 to nDim do pvecTry^[jj]:=pvecGC^[jj] * vFac1 - papvecV^[nHi]^[jj] * vFac2;

vYTry:= func(pDataFunc, pvecTry^);
nNbFuncEval := nNbFuncEval+1;

// If new evaluation is lower than the highest of the last evaluation, keept it
if (vYTry < pavY^[nHi]) then begin
   pavY^[nHi]:= vYTry; // memorise the functions evaluations

   // The gravity center changed, memorise it.
   for jj:=1 to nDim do begin
      pvecGC^[jj]:=pvecGC^[jj] + pvecTry^[jj] - papvecV^[nHi]^[jj];
      papvecV^[nHi]^[jj]:=pvecTry^[jj]; // Replace the vector nHi by the new one
      end;
   end;

TryNewVertex:=vYTry;
end;  // TMinRNSearchRN.TryNewVertex

procedure TMinRNSearchRN.MinimiseSimplex;
//=======================================
var
  nDimp1,
  nHiNext,    // Index of the vertex corresponding of the second highest function evaluation
  ii,jj       : int32;
  vYTry,      // result of one function evaluation
  vErr,       // Difference between highest and lowest function evaluation
  vYSave,     // Save the maximum value of the evaluations
  vSum        // Used to compute the gravity center
              : Float;
  pvecLo,     // vector giving the lower function evaluation
  pvecTemp    // one vector
              : PTypeArrayFloat;
begin
nExitCode   := 0;
nDimp1      :=nDim+1;
nNbFuncEval := 0;

if ((wOptions and 1) = 0) then begin
   // Starting Simplex must be defined.
   pvecLo:=papvecV^[1];
   for ii:=1 to nDim do begin
      pvecTemp:=papvecV^[ii+1];
      for jj:=1 to nDim do pvecTemp^[jj]:=pvecLo^[jj];  // copy the starting vect.

      // the starting Simplex is a cube
      pvecTemp^[ii]:=pvecTemp^[ii] + vSizeEdge;
      end;
   end; // if ((wOptions and 1) = 0) then begin

if ((wOptions and 2) = 0) then begin
   // The function must be evaluated at the starting vectors
   pavY^[1]:=func(pDataFunc, pvecLo^);
   if ((wOptions and 1) = 0) then begin
      for ii:=2 to nDimp1 do pavY^[ii]:=func(pDataFunc, papvecV^[ii]^);
      end;
   end;

// papvecV^[ii] = vector representing one vertex
// compute the center of gravity
for jj:=1 to nDim do begin
   vSum := 0.0;
   for ii:=1 to nDimp1 do vSum:=vSum + papvecV^[ii]^[jj];
   pvecGC^[jj]:=vSum;
   end;

while TRUE do begin
   // Search the min value, max value and next highest values
   nLo := 1;
   if (pavY^[1] > pavY^[2]) then begin
      nHi := 1;
      nHiNext := 2
      end
   else begin
      nHi := 2;
      nHiNext := 1
      end;

   for ii:=1 to nDimp1 do begin
      if (pavY^[ii] < pavY^[nLo]) then nLo := ii;
      if (pavY^[ii] > pavY^[nHi]) then begin
         nHiNext :=nHi;
         nHi     :=ii;
         end
      else if (pavY^[ii] > pavY^[nHiNext]) and (ii <> nHi) then nHiNext := ii;
      end; // for

   // Test if we finished
   vErr:= pavY^[nHi] - pavY^[nLo];
   if (vErr <= vTolAbs + vTolRel * (abs(pavY^[nHi]) + abs(pavY^[nLo]))) then exit; // finished

   if (nNbFuncEval >= nNbFuncEvalMax) then begin
      // Maximum number of iterations has been reached.
      nExitCode:=1;
      exit;
      end;

   // Begin a new iteration
   //"""""""""""""""""""""""
   vYTry := TryNewVertex(-1.0);
   if (vYTry <= pavY^[nLo]) then begin
     // The result is better than the best point, so try another
     // extrapolation by a factor 2.
     vYTry := TryNewVertex(2.0);
     end
   else if (vYTry >= pavY^[nHiNext]) then begin
      // Worse than the second highest value, so try anoter point.
      vYSave :=pavY^[nHi];
      vYTry := TryNewVertex(0.5);

      if (vYTry >= vYSave) then begin
         // No better vertex found, shrink the Simplex near the lowest point
         pvecLo  :=papvecV^[nLo];
         for ii :=1 to nDimp1 do  if (ii <> nLo) then begin
            pvecTemp:=papvecV^[ii];
            for jj := 1 to nDim do begin
               pvecGC^[jj]  :=0.5*(pvecTemp^[jj] + pvecLo^[jj]);
               pvecTemp^[jj]:=pvecGC^[jj];
               end;
            pavY^[ii] := func(pDataFunc, pvecGC^);
            end;

         nNbFuncEval :=nNbFuncEval + nDim;

         // compute the center of gravity
         for jj :=1 to nDim do begin
            vSum := 0.0;
            for ii :=1 to nDimp1 do vSum :=vSum + papvecV^[ii]^[jj];
            pvecGC^[jj] :=vSum;
            end;
         end; // if (vYTry >= vYSave) then begin
      end;
   end;  // while TRUE do begin
end;  // TMinRNSearchRN.MinimiseSimplex

function FuncR1(pData:pointer; vX:Float):Float;
//=============================================
// Minimise the function  Func  allong the line pvecXMin + lambda*pvecDir
var jj : int32;
begin
with PMinRNSearchRN(pData)^ do begin
   for jj:=1 to nDim do pvecX^[jj]:=pvecXMin^[jj] + vX*pvecDir^[jj];

   FuncR1:=Func(pDataFunc, pvecX^);  inc(nNbFuncEval);
   end;
end;  // FuncR1

procedure TMinRNSearchRN.MinimisePowell;
//======================================
// Do the minimisation of the function, using the Powell's methode
var
  oMinR1 : TMinR1SearchR1; // to do minimisation in one direction
  ii,jj,
  nMaxDecrease // index in  papvecV  of the direction of the largest function decrease
               : int32;
  vTest,         // To test if we replace the maximum decrease direction
                 // By the average decrease direction
  vFXMinLast,
  vFDecreaseMax, // Biggest function decrease
  vFXInit,       // Value of vFXMin before line minimisation in the set
                 //  of directions papvecV has been done.
  vFXExtrapol    // = Func(pvecExtrapol)
                 : Float;
  pvecTemp    // one vector
              : PTypeArrayFloat;
begin
oMinR1.Init;

// Data which must be given, before minimisation can start :
oMinR1.func:=@FuncR1;
oMinR1.vXL :=-20*vSizeEdge;
oMinR1.vXR := 20*vSizeEdge; // 20 is quite arbitrary.
oMinR1.vXM :=  0; // Starting guess is at the actual minimum point
oMinR1.vTolAbs   :=vTolAbs;
oMinR1.nNbFuncEvalMax:=nNbFuncEvalMax;
oMinR1.pDataFunc :=@Self;
oMinR1.wOptions:=4; // function is known at pvecMin

if ((wOptions and 1) = 0) then begin
   // Starting directions are define allong the axis
   for ii:=1 to nDim do begin
      pvecTemp:=papvecV^[ii];
      for jj:=1 to nDim do pvecTemp^[jj]:=0.0;

      // papvecV^[ii] = vSizeEdge * (ii_th basis vector)
      pvecTemp^[ii]:=vSizeEdge;
      end;
   end; // if ((wOptions and 1) = 0) then begin

vFXMin := Func(pDataFunc, pvecXMin^);  inc(nNbFuncEval);

// Save the starting point
for jj:=1 to nDim do pvecXMinLast^[jj]:=pvecXMin^[jj];
nNbFuncEval := 0;

while TRUE do begin // Will quite with an exit;
   vFXInit       := vFXMin;
   nMaxDecrease  := 1;    // index of the direction of the biggest decrease
   vFDecreaseMax := 0.0;  // Biggest function decrease

//WriteWin(0, 'loop to minimise');
   // loop over all direction in the set  papvecV
   // To minimise the function along all this directions
   for ii:=1 to nDim do begin
//WriteWin(0, 'minimise in direction' + ITS5(ii));
      for jj:=1 to nDim do pvecDir^[jj] := papvecV^[ii]^[jj];
      vFXMinLast := vFXMin;

      // Minimise the function allong the line pvecMin + lambda*pvecDir
      oMinR1.vXL :=-20*vSizeEdge;
      oMinR1.vXR := 20*vSizeEdge; // 20 is quite arbitrary.
      oMinR1.vXM :=  0; // Starting guess is at the actual minimum point
      oMinR1.vFM:=vFXMin;  // Function is known at the starting point
      oMinR1.wOptions:=4; // function is known at pvecMin
      oMinR1.SearchMinBracket(1.0); // 1.0 = starting step size
      oMinR1.wOptions:=7; // function is known at the 3 starting values
      oMinR1.SearchMinBrent;

      // Result vector
      for jj:=1 to nDim do begin
         pvecXMin^[jj]:=pvecXMin^[jj] + pvecDir^[jj]*oMinR1.vXSol;
//         pvecDir^[jj] :=pvecDir^[jj] * oMinR1.vXM;
         end;

      vFXMin:=oMinR1.vFSol;

      // Memorise the largest decrease
      if (abs(vFXMinLast - vFXMin) > vFDecreaseMax) then begin
         vFDecreaseMax := abs(vFXMinLast - vFXMin);
         nMaxDecrease  := ii;
         end;
      end; // for ii:=...

   // Test if we find that the minimum has been reached near enough
   if (2.0*abs(vFXInit - vFXMin) <= vTolAbs + vTolRel*(abs(vFXInit)+abs(vFXMin))) then begin
      // Finished, the decrease was less than the tolerance
      // pvecXMin = the minimum found
      // vFXMin = value of the function at this minimum
      nExitCode:=0;
      oMinR1.Done;
      exit;
      end;

   if (nNbFuncEval >= nNbFuncEvalMax) then begin
      // Too many functions evaluation, quit the minimisation
      // pvecXMin = the minimum found so far
      // vFXMin = value of the function at this minimum
      nExitCode:=1;
      oMinR1.Done;
      exit;
      end;

   // Construct the extrapolation point,
   // compute the average direction move
   // memorise the old starting point.
   for jj:=1 to nDim do begin
      pvecExtrapol^[jj] := 2.0*pvecXMin^[jj] - pvecXMinLast^[jj];
      pvecDir^[jj]      := pvecXMin^[jj]     - pvecXMinLast^[jj];
      pvecXMinLast^[jj] := pvecXMin^[jj]
      end;

//WriteWin(0, 'extrapolation point');

   // Function value at the extrapolated point
   vFXExtrapol := Func(pDataFunc, pvecExtrapol^);  inc(nNbFuncEval);

   // Test if we keep or replace the maximum decrease direction
   // by the average displacement direction.
   if (vFXExtrapol < vFXInit) then begin
      vTest := 2.0*(vFXInit - 2.0*vFXMin + vFXExtrapol) *
                   sqr(vFXInit - vFXMin - vFDecreaseMax)
             - vFDecreaseMax * sqr(vFXInit - vFXExtrapol);

      if (vTest < 0.0) then begin
         // Move to the minimum in the new direction
//WriteWin(0, 'min search in new direction');
         oMinR1.vXL :=-20*vSizeEdge;
         oMinR1.vXR := 20*vSizeEdge; // 20 is quite arbitrary.
         oMinR1.vXM :=  0; // Starting guess is at the actual minimum point
         oMinR1.vFM :=vFXMin;  // Function is known at the starting point
         oMinR1.wOptions:=4; // function is known at pvecMin
         oMinR1.SearchMinBracket(1.0); // 1.0 = starting step size
         oMinR1.wOptions:=7; // function is known at the 3 starting values
         oMinR1.SearchMinBrent;

         // Result vector
         for jj:=1 to nDim do begin
            pvecXMin^[jj]:=pvecXMin^[jj] + pvecDir^[jj]*oMinR1.vXSol;
            pvecDir^[jj] :=pvecDir^[jj] * oMinR1.vXM;
            end;

         vFXMin:=oMinR1.vFSol;

         // Replace the direction of the maximum decrease, by the new direction
//WriteWin(0, 'replace direction' + ITS5(nMaxDecrease));
         pvecTemp:=papvecV^[nMaxDecrease];
         for jj:=1 to nDim do pvecTemp^[jj]:=pvecDir^[jj];
         end;
      end;
   end; // While
end;  // TMinRNSearchRN.MinimisePowell

end. // unit xMinRNSearchRN;
